from __future__ import annotations

import argparse
import sys

from myver.config import version_from_file, version_to_file
from myver.version import Version


def main(input_args=None):
    """Entry point for the command line utility."""
    args = _parse_args(input_args)

    if args.current:
        version = _current_version(args.config)
        print(version)
    if args.bump:
        version = _current_version(args.config)
        old_version_str = str(version)
        version.bump(args.bump)
        new_version_str = str(version)
        _save_version(version, args.config)
        print(f'{old_version_str}  >>  {new_version_str}')


def _parse_args(args):
    parser = argparse.ArgumentParser(prog='myver')
    parser.add_argument(
        '-b', '--bump',
        help='bump version parts',
        metavar='ARG',
        required=False,
        action='extend',
        nargs='+',
        type=str,
    )
    parser.add_argument(
        '-c', '--current',
        help='get the current version',
        required=False,
        action='store_true',
    )
    parser.add_argument(
        '--config',
        help='config file path',
        required=False,
        default='myver.yml',
        type=str,
    )
    return parser.parse_args(args)


def _current_version(config_path: str):
    try:
        return version_from_file(config_path)
    except FileNotFoundError:
        print(f'File {config_path} not found, it is needed for version '
              f'configuration')
        sys.exit(1)
    except OSError as os_error:
        print(f'OS error occurred trying to open {config_path}')
        print(os_error.strerror)
        sys.exit(1)


def _save_version(version: Version, config_path: str):
    try:
        return version_to_file(config_path, version)
    except FileNotFoundError:
        print(f'File {config_path} not found, it is needed for version '
              f'configuration')
        sys.exit(1)
    except OSError as os_error:
        print(f'OS error occurred trying to open {config_path}')
        print(os_error.strerror)
        sys.exit(1)



from __future__ import annotations

import abc
from typing import Optional, Union

from myver.error import ConfigError, BumpError


class Part(abc.ABC):
    """The base class for a version part.

    :param key: The unique key of the part. This is used to set dict
        keys for collections of parts.
    :param value: The actual value of the part.
    :param requires: Another part that this part requires. This means
        that the required part will need to be set if this part is set.
    """

    def __init__(self,
                 key: str,
                 value: Optional[Union[str, int]],
                 requires: Optional[str] = None,
                 prefix: Optional[str] = None,
                 child: Optional[Part] = None,
                 parent: Optional[Part] = None):
        self.prefix: str = prefix or ''
        self.key: str = key
        self.value: Optional[Union[str, int]] = value
        self.requires: Optional[str] = requires
        self._child: Optional[Part] = None
        self._parent: Optional[Part] = None
        self.child = child
        self.parent = parent

    @abc.abstractmethod
    def next_value(self, value_override: str = None):
        """Set current value to the next part value.

        :param value_override: Manual override for the bumped value.
        """

    @property
    @abc.abstractmethod
    def start(self) -> Union[str, int]:
        """Get the start value in a usable form (i.e. not None)"""

    @start.setter
    @abc.abstractmethod
    def start(self, new_start: Union[str, int]):
        """Set the start value"""

    @property
    def child(self) -> Optional[Part]:
        return self._child

    @child.setter
    def child(self, part: Optional[Part]):
        self._child = part
        if self._child and not self._child.parent:
            self._child.parent = self

    @property
    def parent(self) -> Optional[Part]:
        return self._parent

    @parent.setter
    def parent(self, part: Optional[Part]):
        self._parent = part
        if self._parent and not self._parent.child:
            self._parent.child = self

    def is_set(self) -> bool:
        """Checks if the part's value is not None."""
        return self.value is not None

    def bump(self, bump_args: list[str] = None, value_override: str = None):
        """Bump this part's value.

        :param bump_args: The bump arguments.
        :param value_override: Manual override for the bumped value.
        """
        bump_args = bump_args or []
        self.next_value(value_override)
        if self.child:
            self.child.reset(bump_args)

    def reset(self, bump_args: list[str] = None):
        """Reset part value to the start value.

        Resetting the part to the start value will also make a recursive
        call to its child, resetting their values too.

        :param bump_args: The keys that are being bumped.
        """
        bump_args = bump_args or []

        # If this part is required and it's in the bump keys, we want to
        # skip this step so that we do not get a double bump.
        if self.is_required() and self.key not in bump_args:
            self.value = self.start
        else:
            self.value = None

        if self.child:
            self.child.reset()

    def is_required(self) -> bool:
        """Checks if this part is required by any parents."""
        return self._parent_requires(self.key)

    def _parent_requires(self, key: str) -> bool:
        """Check if a part is required based on its key.

        This does a recursive call up to all of the parents until it
        reaches the final parent to check if any of them require the
        part specified.

        :param key: The key of the part that you want to check.
        """
        if self.parent is not None:
            if self.parent.requires == key and self.parent.is_set():
                return True
            else:
                return self.parent._parent_requires(key)
        return False

    def __str__(self):
        return f'{self.prefix}{self.value}'

    def __eq__(self, other: Part) -> bool:
        return (self.key == other.key) and (self.value == other.value)


class IdentifierPart(Part):
    """An identifier part.

    :param strings: List of valid strings that can be used as an
        identifier for this part.
    :param start: The starting value of the part. This is used when the
        part goes out of a null state, or is reset to its original
        state. If this is specified it must be a string that is in the
        `self.strings` list.
    """

    def __init__(self,
                 key: str,
                 value: Optional[str],
                 strings: list[str],
                 requires: Optional[str] = None,
                 prefix: Optional[str] = None,
                 child: Optional[Part] = None,
                 parent: Optional[Part] = None,
                 start: str = None):
        super().__init__(key, value, requires, prefix, child, parent)
        self._strings: list[str] = strings
        self._start: Optional[str] = start
        self.strings = strings
        self.start = start

    @property
    def strings(self):
        return self._strings

    @strings.setter
    def strings(self, new_strings: list[str]):
        self._validate_strings(new_strings)
        self._strings: list[str] = new_strings

    @property
    def start(self) -> str:
        return self._start or self.strings[0]

    @start.setter
    def start(self, new_start: str):
        self._validate_start(new_start)
        self._start = new_start

    def next_value(self, value_override: str = None):
        if value_override is not None:
            if value_override not in self.strings:
                raise BumpError(
                    f'Cannot set value override `{value_override}`, it must '
                    f'be in the strings list of part `{self.key}`')
            else:
                self.value = value_override
        elif self.is_set():
            current_index = self.strings.index(self.value)
            next_index = current_index + 1

            if next_index < len(self.strings):
                self.value = self.strings[next_index]
            else:
                self.value = None
        else:
            self.value = self.start

    def _validate_start(self, start: Optional[str]):
        if start is not None and start not in self.strings:
            raise ConfigError(
                f'Part `{self.key}` has an `identifier.start` value that is '
                f'not in the `identifier.strings` list')

    def _validate_strings(self, strings: list[str]):
        if not len(strings) > 0:
            raise ConfigError(
                f'Part `{self.key}` has an `identifier.strings` has an empty '
                f'list, the list must have at least one string')


class NumberPart(Part):
    """A number part.

    :param label_suffix: String to use for separating the label and the
        number.
    :param start: Starting value of the part. This is used when the part
        goes out of a null state, or is reset to its original state.
    :param show_start: If true, the start value will be shown in the
        version. If false, then the start value wont be shown although
        the next value (after a bump) will be shown.
    """

    def __init__(self,
                 key: str,
                 value: Optional[int],
                 requires: Optional[str] = None,
                 prefix: Optional[str] = None,
                 child: Optional[Part] = None,
                 parent: Optional[Part] = None,
                 label: Optional[str] = None,
                 label_suffix: Optional[str] = None,
                 start: int = None,
                 show_start: bool = None):
        super().__init__(key, value, requires, prefix, child, parent)
        self.label: Optional[str] = label or ''
        self.label_suffix: Optional[str] = label_suffix or ''
        if show_start is None:
            self.show_start: bool = True
        else:
            self.show_start: bool = show_start
        self._start: Optional[int] = start
        self.start = start

    @property
    def start(self) -> int:
        return self._start or 0

    @start.setter
    def start(self, new_start: int):
        self._validate_start(new_start)
        self._start = new_start

    def next_value(self, value_override: str = None):
        if value_override is not None:
            try:
                int(value_override)
            except ValueError:
                raise BumpError(
                    f'Value override `{value_override}` must be a number for '
                    f'part `{self.key}`')

            if int(value_override) < 0:
                raise BumpError(
                    f'Cannot set value override to a negative number '
                    f'`{value_override}` for part `{self.key}`')
            self.value = int(value_override)
        elif self.is_set():
            self.value = self.value + 1
        else:
            self.value = self.start

    def _validate_start(self, start: Optional[int]):
        if start is not None:
            try:
                int(start)
            except ValueError:
                raise ConfigError(
                    f'Part `{self.key}` has an invalid value for its '
                    f'`number.start` attribute, it must be an integer')

        if start is not None and start < 0:
            raise ConfigError(
                f'Part `{self.key}` has an negative value for its '
                f'`number.start` attribute, it must be positive')

    def __str__(self):
        if self.value == self.start and not self.show_start:
            return f'{self.prefix}{self.label}'
        return f'{self.prefix}{self.label}{self.label_suffix}{self.value}'



from __future__ import annotations

from myver.error import ConfigError, BumpError
from myver.part import Part


class Version:
    """Represents the version itself.

    This is the top level class for a version. It contains the groups
    of parts and this is where the version operations are performed.

    :param parts: The list of parts in the version.
    """

    def __init__(self, parts: list[Part] = None):
        self._parts: list[Part] = parts or list()
        self.parts = parts or list()

    @property
    def parts(self) -> list[Part]:
        return self._parts

    @parts.setter
    def parts(self, new_parts: list[Part]):
        """Sets the parts list.

        :param new_parts: The parts to set.
        :raise KeyConflictError: A part key appears 2 or more times in
            the list.
        """
        validate_keys(new_parts)
        validate_requires(new_parts)
        self._parts = new_parts
        set_relationships(self._parts)

    def bump(self, args: list[str]):
        """Bump the version based on bumping args.

        :param args: The list of part keys to bump. An arg may have a
            key value pair with the syntax of `<key>=<value>`.
        :raise BumpError: When the bumping fails.
        """
        for arg in args:
            if arg.count('=') > 1:
                raise BumpError(
                    f'The bump arg `{arg}` contains more than 1 `=` symbol, '
                    f'there can only be 0 or 1 `=` symbol in the bump arg')
            elif arg.count('=') == 1:
                key = arg.split('=')[0]
                value_override = arg.split('=')[1]
                self.part(key).bump(args, value_override)
            else:
                self.part(arg).bump(args)

    def part(self, key: str) -> Part:
        """Gets a part based on its key.

        :param key: The key of the part you are getting.
        :raise KeyError: If no part has the key provided.
        """
        for part in self._parts:
            if part.key == key:
                return part
        raise KeyError(key)

    def __eq__(self, other: Version):
        for this_part, other_part in zip(self.parts, other.parts):
            if not this_part == other_part:
                return False
        return True

    def __str__(self):
        version_str = ''

        for part in self._parts:
            if part.is_set():
                version_str += str(part)

        return version_str


def validate_requires(parts: list[Part]):
    """Validates that parts require other valid parts.

    :raise ConfigError: If a part requires itself or a part that does
        not exist.
    """
    keys = [p.key for p in parts] or []
    for part in parts:
        if not part.requires:
            continue

        if part.requires == part.key:
            raise ConfigError(
                f'Part `{part.key}` has a `requires` key that is'
                f'referencing itself, it must reference another part')
        if part.requires not in keys:
            raise ConfigError(
                f'Part `{part.key}` has a `requires` key '
                f'"{part.requires}" that does not exist, it must be a '
                f'valid key of another part')


def validate_keys(parts: list[Part]):
    """Validates that they keys are unique.

    :raise ConfigError: If two or more parts with the same key.
    """
    keys = [p.key for p in parts] or []
    for key in keys:
        if keys.count(key) > 1:
            raise ConfigError(
                f'Key "{key}" is configured on more than one part, all '
                f'parts must have a unique key')


def set_relationships(parts: list[Part]):
    """Sets the parent-child relationships between a list of parts.

    :param parts: The parts to set the relationships for.
    """
    for i in range(len(parts)):
        if i < len(parts) - 1:
            parts[i].child = parts[i + 1]



from __future__ import annotations

import ruamel.yaml

from myver.error import ConfigError
from myver.part import Part, IdentifierPart, NumberPart
from myver.version import Version


def dict_from_file(path: str) -> dict:
    """Gets the dict config from a file.

    The default file path is `myver.yml`, which is a relative path. This
    path can be overridden by using the `path` arg.

    :param path: The path to the myver config file.
    :raise FileNotFoundError: If the file does not exist.
    :raise OSError: For other errors when accessing the file.
    """
    with open(path, 'r') as file:
        yaml = ruamel.yaml.YAML()
        config_dict = yaml.load(file)
    return config_dict


def version_to_file(path: str, version: Version):
    """Syncs a version to a yaml file.

    This will only sync the part values to the file, no other
    configuration of the file will change. This will mean that the parts
    in the version will need to have perfect 1:1 corresponding keys for
    each part and within the yaml file and in the `version` param. This
    also means that the yaml file must have existing configuration
    details for each part in the `version` param.

    :param path: The path of the yaml file.
    :param version: The version to sync to the yaml file.
    :raise FileNotFoundError: If the file does not exist.
    :raise OSError: For other errors when accessing the file.
    :raise ConfigError: When the yaml file does not have a 1:1 of keys
        for parts compared to the `version` param.
    """
    try:
        _update_file_values(path, version)
    except KeyError as key_error:
        key = key_error.args[0]
        raise ConfigError(
            f'You must have the required attribute `{key}` configured')


def _update_file_values(path: str, version: Version):
    """Update config file part based on `version` object.

    :param path: The path to the config file.
    :param version: The version to persist to the file.
    """
    update_map = _get_value_update_map(path, version)

    # We need the lines to rewrite the changes later.
    with open(path, 'r') as file:
        lines = file.readlines()

    with open(path, 'w') as file:
        for index, value in update_map.items():
            indent_length = len(lines[index]) - len(lines[index].lstrip())
            lines[index] = f'{" " * indent_length}value: {value}\n'

        file.writelines(lines)


def _get_value_update_map(path: str, version: Version) -> dict[int, str]:
    """Get an update map for version values in a config file.

    :param path: The path to the config file.
    :param version: The version to persist to the file.
    :return: A dict with each key in the dict represents the index in
        the lines list to update (based on index counter starting at 0).
        The value of each entry in the dict represents a part's value to
        be updated at the given line index.
    """
    config_dict = dict_from_file(path)
    update_map = dict()

    with open(path, 'r') as file:
        lines = file.readlines()

        for key in config_dict['parts'].keys():
            # We want to start at the part's key so that the next
            # `value` node is guaranteed to be the part's `value` node.
            from_index = config_dict['parts'][key].lc.line
            # Note, it's an index (as related to list indexes) and not
            # a line number.
            line_index = _find_value_node_index(lines, from_index)
            part = version.part(key)

            if part.value is None:
                update_map[line_index] = 'null'
            else:
                update_map[line_index] = part.value

    return update_map


def _find_value_node_index(lines: list[str], from_index: int) -> int:
    """Find the line index for the `value` node.

    :param lines: The lines to read from in order to get the index.
    :param from_index: The index to start searching from.
    :return: The index of where the `value` node is.
    """
    for i, line in enumerate(lines[from_index:]):
        if line.lstrip().startswith('value:'):
            return i + from_index


def version_from_file(path: str) -> Version:
    """Construct version from a config dict.

    :param path: The path of the yaml file.
    :raise ConfigError: If the configuration is invalid.
    :return: The version.
    """
    config_dict = dict_from_file(path)
    return version_from_dict(config_dict)


def version_from_dict(config_dict: dict) -> Version:
    """Construct version from a config dict.

    :param config_dict: The dict with raw version config data.
    :raise ConfigError: If the configuration dict is invalid.
    :return: The version.
    """
    try:
        parts: list[Part] = []
        for part_key, part_dict in config_dict['parts'].items():
            parts.append(part_from_dict(part_key, part_dict))
        return Version(parts)
    except KeyError as key_error:
        key = key_error.args[0]
        raise ConfigError(
            f'You must have the required attribute `{key}` configured')


def part_from_dict(key: str, config_dict: dict) -> Part:
    """Construct part from a config dict.

    :param key: The part's key.
    :param config_dict: The dict with raw part config data.
    :raise ConfigError: If the configuration dict is invalid.
    :raise KeyError: If the config is missing required attributes.
    :return: The version part.
    """
    if config_dict.get('identifier') and config_dict.get('number'):
        raise ConfigError(
            f'Part `{key}` cannot be an identifier and number at the '
            f'same time. Configure either `number` or `identifier` '
            f'attribute')
    elif config_dict.get('identifier'):
        return IdentifierPart(
            key=key,
            value=config_dict['value'],
            strings=config_dict['identifier']['strings'],
            requires=config_dict.get('requires'),
            prefix=config_dict.get('prefix'),
            start=config_dict['identifier'].get('start'))
    elif config_dict.get('number'):
        return NumberPart(
            key=key,
            value=config_dict['value'],
            requires=config_dict.get('requires'),
            prefix=config_dict.get('prefix'),
            label=config_dict['number'].get('label'),
            label_suffix=config_dict['number'].get('label-suffix'),
            start=config_dict['number'].get('start'),
            show_start=config_dict['number'].get('show-start'))
    else:
        # Default if no type configuration is specified.
        return NumberPart(
            key=key,
            value=config_dict['value'],
            requires=config_dict.get('requires'),
            prefix=config_dict.get('prefix'))



class Error(Exception):
    """Configuration is invalid."""
    def __init__(self, message: str):
        self.message: str = message
        super().__init__()

    def __str__(self):
        return self.message


class ConfigError(Error):
    """Configuration is invalid."""


class BumpError(Error):
    """Bumping a version or part failed."""
